from Crypto.Cipher import AES
import hashlib
import os

class CryptoError(Exception):
    pass

def raw_encrypt(key, data, chunk=1024*32):
    key = stretch_key(key, algo=hashlib.sha256)

    # Create random 16 byte initialization vector.  
    iv = os.urandom(16)
    size = len(data)

    # Using cipher block chain mode; effectively means each block of plaintext
    # is XORed with the previous block's ciphertext. This means each
    # ciphertext block is unique up to each point.  
    crypter = AES.new(key, AES.MODE_CBC, iv)
    encrypted = iv + crypter.encrypt(SUCCESS_INDICATOR)

    for x in xrange(size // chunk + 1):
        td = data[x * chunk:(x + 1) * chunk]
        td = pad(td, 16, " ")
        encrypted += crypter.encrypt(td)

    return encrypted

def raw_decrypt(key, data, chunk=1024*32):
    key = stretch_key(key, algo=hashlib.sha256)
    retval = ""

    # Retrieve the initialization vector.  
    iv = data[:16]
    data = data[16:]
    crypter = AES.new(key, AES.MODE_CBC, iv)

    suc = crypter.decrypt(data[:len(SUCCESS_INDICATOR)])
    data = data[len(SUCCESS_INDICATOR):]
    
    # If this condition is true, we can assume with extremely high
    # probability that decryption was successful.  
    if suc != SUCCESS_INDICATOR:
        raise CryptoError("Decryption failed -- bad key or input data?")

    while 1:
        c = data[:chunk]
        data = data[chunk:]
        if not len(c):
            break
        retval += crypter.decrypt(pad(c, 16, " "))
    
    return retval

def encrypt(key, data, outfile, chunk=1024*32, file_permissions=0600):
    """Encrypts some data and writes it to a file 'outfile'.
        key -- the encryption key to use.
        data -- the data to encrypt.
        outfile -- the name of the file to output to.
        chunk (optional) -- the number of bytes to operate on in one round.
    """
    with open(outfile, "wb") as f:
        # Write the initialization vector to the file.  
        f.write(raw_encrypt(key, data, chunk))

    # Assign file permissions.  
    os.chmod(outfile, file_permissions)

def decrypt(key, infile, chunk=1024*32):
    """Decrypts some data from an 'infile' and returns the plaintext.
        key -- the encryption key to use.
        infile -- the name of the file to take input from.
        chunk (optional) -- the number of bytes to operate on in one round.
    """
    data = ""
    with open(infile, "rb") as f:
        data = f.read()
    
    return raw_decrypt(key, data, chunk)

def pad(data, bs, pad_char):
    """Pads a string. This is needed to transform a string of n bytes long
    to one whose length is equal to the specified 'bs' argument.
    """
    return data + (bs - len(data) % bs) * pad_char

def stretch_key(key, rounds=65536, salt="", algo=hashlib.sha256):
    """Stretches an encryption key by a default of 65536 rounds."""
    stretched = key
    for _ in xrange(rounds):
        stretched = algo(stretched + key + salt).digest()
    return stretched

SUCCESS_INDICATOR = pad("SUCCESS", 16, " ")
