from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import crypto
import getpass
import glob
import hashlib
import os
import pickle
import tarfile

DIRECTORY = os.path.expanduser("~") + os.path.sep + ".jpass" + os.path.sep

class Entry(object):
    
    def __init__(self, name, username="", passwd="", url="", notes=""):
        self.name = name
        self.username = username
        self.passwd = passwd
        self.url = url
        self.notes = notes
    
    def write(self, key, path):
        serialized = pickle.dumps(self)
        ciphertext = crypto.encrypt(key, serialized, path)

    def __str__(self):
        s = "== %s ==\n" % self.name
        s += "Username: %s\n" % self.username if self.username else ""
        s += "Password: %s\n" % self.passwd if self.passwd else ""
        s += "URL: %s\n" % self.url if self.url else ""
        s += "Notes: %s\n" % self.notes if self.notes else ""
        return s

def read(key, path):
    plaintext = crypto.decrypt(key, path)
    obj = pickle.loads(plaintext)
    return obj

def init():
    if not os.path.isdir(DIRECTORY):
        os.mkdir(DIRECTORY)

def backup_db():
    files = os.listdir(DIRECTORY)
    if len(files) == 0:
        print "There is nothing in ~/.jpass to backup!"
        return

    print "Creating a backup of the ~/.jpass directory..."
    with tarfile.open("jpass.tar.gz", "w:gz") as tar:
        for name in files:
            print name
            tar.addfile(tarfile.TarInfo(name), file(DIRECTORY + name))
    print "Done!"

def edit_entry(name):
    path = DIRECTORY + name
    if os.path.isfile(path):

        try:
            key = getpass.getpass("Key: ")
            print "\n",

            entry = read(key, path)
            print "As it stands, %s contains the following data:\n" % path
            print entry
            
            while 1:
                print "Choose a field to edit (enter nothing when done):"
                i = raw_input("<username/password/url/notes> ").lower()

                if not i: break
                if i == "username":
                    entry.username = raw_input("Username: ")
                elif i == "password":
                    entry.passwd = raw_input("Password: ")
                elif i == "url":
                    entry.url = raw_input("URL: ")
                elif i == "notes":
                    entry.notes = raw_input("Notes: ")
                else:
                    print "I can't edit that."
                
                print "\n",
                print "%s now contains the following data:\n" % path
                print entry
            
            entry.write(key, path)
            print "Done!"

        except crypto.CryptoError:
            print "Decryption failed. Bad key or input data?"

    else:
        print "Entry %s doesn't exist, so you can't edit it." % path

def list_entries():
    entries = glob.glob(DIRECTORY + "*.jpass")
    if entries:
        print "You have the following entry files in ~/.jpass:"
        print "\n".join(e.rpartition("/")[-1] for e in entries)
    else:
        print "You have no entry files in ~/.jpass."

def read_entry(name):
    path = DIRECTORY + name
    if os.path.isfile(path):
        key = getpass.getpass("Key: ")
        try:
            print "\n",
            print read(key, path)
            print "Done!"
        except crypto.CryptoError:
            print "Decryption failed. Bad key or input data?"
    else:
        print "Entry %s doesn't exist, so you can't read it." % path

def remove_entry(name):
    path = DIRECTORY + name
    if os.path.isfile(path):
        os.remove(path)
        print "Successfully removed %s." % path
        print "Done!"
    else:
        print "Entry %s doesn't exist, so you can't remove it." % path

def restore_db():
    name = ""
    while 1:
        name = raw_input("Enter the path to the DB backup: ")
        if os.path.isfile(name): break
        else: print "Please enter a valid path."

    print "Extracting entries from %s to %s..." % (name, DIRECTORY)
    try:
        tar = tarfile.open(name)
        tar.extractall(path=DIRECTORY)
        tar.close()
    except tarfile.ReadError:
        print "Could not read tar file."
        exit(1)
    print "Done!"

def store_entry(name):
    path = DIRECTORY + name
    if not os.path.isfile(path):
        username = raw_input("Username: ")
        password = raw_input("Password: ")
        url = raw_input("URL: ")
        notes = raw_input("Notes: ")
        print "\n",

        e = Entry(name, username, password, url, notes)
        key = getpass.getpass("Key: ")
        while key != getpass.getpass("Confirm: "):
            print "Passwords did not match!\n"
            key = getpass.getpass("Key: ")

        e.write(key, path)
        print "\n",
        print "Done!"
    else:
        print "%s is already an entry!" % path
        print "You can edit it or delete it using the edit or remove tasks."

def main():
    ap = ArgumentParser(description="Manage your passwords.",
                        formatter_class=ArgumentDefaultsHelpFormatter)
    ap.add_argument("task", help="The task to perform.")
    ap.add_argument("--entry", "-e", help="The entry to work on.")
    args = ap.parse_args()

    tasks = {
        
        "backup" : backup_db,
        "copy" : backup_db,

        "change" : edit_entry,
        "edit" : edit_entry,
        "modify" : edit_entry,
        "tweak" : edit_entry,

        "list" : list_entries,
        "ls" : list_entries,

        "look" : read_entry,
        "read" : read_entry,
        "see" : read_entry,
        "view" : read_entry,

        "delete" : remove_entry,
        "remove" : remove_entry,
        "rm" : remove_entry,

        "restore" : restore_db,

        "create" : store_entry,
        "make" : store_entry,
        "store" : store_entry,
    }

    init()

    task = args.task.lower()
    try:
        tasks[task](args.entry + ".jpass")
    except TypeError:
        # If the function doesn't want a name to work with,
        # don't give it one.
        # TODO: perhaps a better way to do this...  
        tasks[task]()
    except KeyError:
        # If the task that the user has given doesn't exist, tell
        # them.  
        print "No such task '%s'." % args.task

if __name__ == "__main__":
    try:
        main()
    except (EOFError, KeyboardInterrupt):
        print "\n",
        print "Interrupt caught; exiting. Have a nice day!"
